from __future__ import print_function
import numpy as np
import fc_bench
from fc_bench.bench import bData

def all_benchs(**kwargs):
  nbruns=kwargs.pop('nbruns',5)
  debug=kwargs.pop('debug',False)
  Lb= [ bench_aXpbY_all, 
        bench_plus_all,bench_minus_all,bench_times_all,bench_rdivide_all,bench_ldivide_all,
        bench_mtimes_all, bench_det,
        bench_PALU, bench_cholesky, bench_pLUinplace,
        bench_linsolve_diag_all,bench_linsolve_tril_all,bench_linsolve_triu_all]
  if debug:
    kwargs['nbruns']=1
    kwargs['c']=1
  else:
    kwargs['nbruns']=nbruns
  for bch in Lb:
    print('='*60)
    print(' Running %s.%s function'%(bch.__module__,bch.__name__))
    print('='*60)
    bch(**kwargs)
    print(' End of %s.%s function '%(bch.__module__,bch.__name__))
    print('='*60)

TypeArray=['array3d','array2d','array1d','scalar']

def set_datas(N,verbose,**kwargs):
  d=kwargs.pop('d',[2,2]) 
  Type=kwargs.pop('Type','array3D').lower()  # LOWER CASE
  Print=kwargs.pop('Print',lambda s: print(s))
  genfun=kwargs.pop('genfun',lambda *S: np.random.randn(*S))
  assert Type in ['array3d','array2d','array1d','scalar']
  
  if Type=='array3d':
    A=genfun(N,d[0],d[1])
  if Type=='array2d':
    A=genfun(d[0],d[1])  
  if Type=='array1d':
    A=genfun(N)
  if Type=='scalar':
    A=genfun()
  if verbose:
    if Type=='array3d':
      Print('# -> 3D-array with (N,m,n)=(%d,%d,%d),  shape=%s'%(N,d[0],d[1],str(A.shape)))
    if Type=='array2d':
      Print('# -> 2D-array with (m,n)=(%d,%d),  shape=%s'%(d[0],d[1],str(A.shape)))
    if Type=='array1d':
      Print('# -> 1D-array with N=%d,  shape=%s'%(N,str(A.shape)))
    if Type=='scalar':
      Print('# -> scalar with isscalar=%s'%(str(np.isscalar(A))))
  
  bDs=[bData('{:>8}'.format('N'),N,'{:8}')]
  return (A,bDs)

def set_mtimes(N,verbose,**kwargs):
  Print=kwargs.get('Print',lambda s: print(s))
  dl=kwargs.pop('dl',[2,2]) 
  dr=kwargs.pop('dr',[2,2]) 
  ltype=kwargs.pop('ltype','array3D').lower()
  rtype=kwargs.pop('rtype','array3D').lower() 
  assert ltype in ['array3d','array2d']
  assert rtype in ['array3d','array2d']
  assert ltype=='array3d' or rtype=='array3d'
  if verbose:
    Print('# Left input is:')
  A,bdA=set_datas(N,verbose,Type=ltype,d=dl,**kwargs)
  if verbose:
    Print('# Right input is:')
  B,bdB=set_datas(N,verbose,Type=rtype,d=dr,**kwargs)
  return ((A,B),bdA)

def set_aXpbY(N,verbose,**kwargs):
  Print=kwargs.get('Print',lambda s: print(s))
  #d=kwargs.pop('d',[2,2]) 
  ltype=kwargs.pop('ltype','array3D').lower()
  rtype=kwargs.pop('rtype','array3D').lower() 
  assert ltype in ['array3d','array2d','array1d','scalar']
  assert rtype in ['array3d','array2d','array1d','scalar']
  assert ltype=='array3d' or rtype=='array3d'
  if verbose:
    Print('# 1st input parameter is:')
  a,bd=set_datas(N,verbose,Type='scalar',**kwargs)
  if verbose:
    Print('# 2nd input parameter is:')
  X,bdX=set_datas(N,verbose,Type=ltype,**kwargs)
  if verbose:
    Print('# 3rd input parameter is:')
  b,bd=set_datas(N,verbose,Type='scalar',**kwargs)
  if verbose:
    Print('# 4nd input parameter is:')
  Y,bdB=set_datas(N,verbose,Type=rtype,**kwargs)
  bDs=[bData('{:>8}'.format('N'),N,'{:8}')]
  return ((a,X,b,Y),bDs)

def set_linsolve_diag(N,verbose,**kwargs):
  from fc_linalg3D import random,mtimes,aXpbY
  Print=kwargs.get('Print',lambda s: print(s))
  d=kwargs.pop('d',2)
  m=kwargs.pop('m',1)
  rtype=kwargs.pop('rtype','array3D').lower() 
  assert rtype in ['array3d','array2d']
  A=random.randdiag(N,d,a=1,b=2)
  if verbose:
    Print('# 1st input parameter is:')
    Print('# -> 3D-array with (N,m,n)=(%d,%d,%d),  shape=%s'%(N,d,d,str(A.shape)))
  if verbose:
    Print('# 2nd input parameter is:')
  B,bdB=set_datas(N,verbose,d=[d,m],Type=rtype,**kwargs)
  bDs=[bData('{:>8}'.format('N'),N,'{:8}')]
  Error=lambda X: np.max(np.abs(aXpbY(1,mtimes(A,X),-1,B)))
  return ((A,B,Error),bDs)

def set_linsolve_tril(N,verbose,**kwargs):
  from fc_linalg3D import random
  from fc_linalg3D import mtimes,aXpbY
  Print=kwargs.get('Print',lambda s: print(s))
  d=kwargs.pop('d',2)
  m=kwargs.pop('m',1)
  rtype=kwargs.pop('rtype','array3D').lower() 
  assert rtype in ['array3d','array2d']
  A=random.randtril(N,d,a=1,b=2)
  if verbose:
    Print('# 1st input parameter is:')
    Print('# -> 3D-array with (N,m,n)=(%d,%d,%d),  shape=%s'%(N,d,d,str(A.shape)))
  if verbose:
    Print('# 2nd input parameter is:')
  B,bdB=set_datas(N,verbose,d=[d,m],Type=rtype,**kwargs)
  bDs=[bData('{:>8}'.format('N'),N,'{:8}')]
  Error=lambda X: np.max(np.abs(aXpbY(1,mtimes(A,X),-1,B)))
  return ((A,B,Error),bDs)

def set_linsolve_triu(N,verbose,**kwargs):
  from fc_linalg3D import random
  from fc_linalg3D import mtimes,aXpbY
  Print=kwargs.get('Print',lambda s: print(s))
  d=kwargs.pop('d',2)
  m=kwargs.pop('m',1)
  rtype=kwargs.pop('rtype','array3D').lower() 
  assert rtype in ['array3d','array2d']
  A=random.randtriu(N,d,a=1,b=2)
  if verbose:
    Print('# 1st input parameter is:')
    Print('# -> 3D-array with (N,m,n)=(%d,%d,%d),  shape=%s'%(N,d,d,str(A.shape)))
  if verbose:
    Print('# 2nd input parameter is:')
  B,bdB=set_datas(N,verbose,d=[d,m],Type=rtype,**kwargs)
  bDs=[bData('{:>8}'.format('N'),N,'{:8}')]
  Error=lambda X: np.max(np.abs(aXpbY(1,mtimes(A,X),-1,B)))
  return ((A,B,Error),bDs)

def set_linsolve_LU(N,verbose,**kwargs):
  from fc_linalg3D import random
  from fc_linalg3D import mtimes,aXpbY
  Print=kwargs.get('Print',lambda s: print(s))
  d=kwargs.pop('d',2)
  m=kwargs.pop('m',1)
  rtype=kwargs.pop('rtype','array3D').lower() 
  assert rtype in ['array3d','array2d']
  A=mtimes(random.randtril(N,d,a=1,b=2),random.randtriu(N,d,a=-2,b=-1))
  if verbose:
    Print('# 1st input parameter is:')
    Print('# -> 3D-array with (N,m,n)=(%d,%d,%d),  shape=%s'%(N,d,d,str(A.shape)))
  if verbose:
    Print('# 2nd input parameter is:')
  B,bdB=set_datas(N,verbose,d=[d,m],Type=rtype,**kwargs)
  bDs=[bData('{:>8}'.format('N'),N,'{:8}')]
  Error=lambda X: np.max(np.abs(aXpbY(1,mtimes(A,X),-1,B)))
  return ((A,B,Error),bDs)

def set_linsolve_cholesky(N,verbose,**kwargs):
  from fc_linalg3D import random
  from fc_linalg3D import mtimes,aXpbY
  Print=kwargs.get('Print',lambda s: print(s))
  d=kwargs.pop('d',2)
  m=kwargs.pop('m',1)
  rtype=kwargs.pop('rtype','array3D').lower() 
  assert rtype in ['array3d','array2d']
  A=random.randnsympd(N,d)
  if verbose:
    Print('# 1st input parameter is:')
    Print('# -> 3D-array with (N,m,n)=(%d,%d,%d),  shape=%s'%(N,d,d,str(A.shape)))
  if verbose:
    Print('# 2nd input parameter is:')
  B,bdB=set_datas(N,verbose,d=[d,m],Type=rtype,**kwargs)
  bDs=[bData('{:>8}'.format('N'),N,'{:8}')]
  Error=lambda X: np.max(np.abs(aXpbY(1,mtimes(A,X),-1,B)))
  return ((A,B,Error),bDs)

def set_ebye(N,verbose,**kwargs):
  d=kwargs.pop('d',[2,2]) 
  ltype=kwargs.pop('ltype','array3D').lower()
  rtype=kwargs.pop('rtype','array3D').lower() 
  assert ltype in ['array3d','array2d','array1d','scalar']
  assert rtype in ['array3d','array2d','array1d','scalar']
  assert ltype=='array3d' or rtype=='array3d'
  
  A,bdA=set_datas(N,verbose,Type=ltype,d=d,**kwargs)
  B,bdB=set_datas(N,verbose,Type=rtype,d=d,**kwargs)
  return ((A,B),bdA)

def set_sympd(N,verbose,**kwargs):
  d=kwargs.pop('d',2) 
  Print=kwargs.pop('Print',lambda s: print(s))
  from fc_linalg3D.random import randnsympd
  A=randnsympd(N,d)
  if verbose:
    Print('# Symmetric Positive Definite matrices')
    Print('# -> 3D-array with (N,m,n)=(%d,%d,%d),  shape=%s'%(N,d,d,str(A.shape)))
  
  bDs=[bData('{:>8}'.format('N'),N,'{:8}')]
  return (A,bDs)

def error_linsolve(A,b,x):
  from fc_linalg3D.mtimes import mtimes_Vec as mtimes
  from fc_linalg3D.ebye import aXpbY_Vec as aXpbY
  return np.max(abs(aXpbY(1,mtimes(A,x),-1,b)))
  
def set_linsolve(N,verbose,**kwargs):
  from fc_linalg3D.random import randnsympd
  from fc_linalg3D.mtimes import mtimes_Vec as mtimes
  Print=kwargs.get('Print',lambda s: print(s))
  dl=kwargs.pop('dl',[2,2]) 
  dr=kwargs.pop('dr',[2,1]) 
  assert dl[1]==dr[0]
  rtype=kwargs.pop('rtype','array3D').lower() 
  assert rtype in ['array3d','array2d']
  if verbose:
    Print('# Left input is:')
  A,bdA=set_datas(N,verbose,Type='array3D',d=dl,genfun=lambda N,m,n: randnsympd(N,m),**kwargs)
  if verbose:
    Print('# Right input is:')
  B,bdB=set_datas(N,verbose,Type=rtype,d=dr,**kwargs)
  error=lambda X: np.max(abs(mtimes(A,X)-B))
  return ((A,B,error),bdA)

def set_cholesky(N,verbose,**kwargs):
  from fc_linalg3D.random import randnsympd
  from fc_linalg3D import ctranspose,mtimes
  d=kwargs.pop('d',4) 
  A,bdA=set_datas(N,verbose,Type='array3D',d=[d,d],genfun=lambda N,m,n: randnsympd(N,m),**kwargs)
  error=lambda X: np.max(abs(mtimes(X,ctranspose(X))-A))
  return ((A,error),bdA)

def error_pALU(A,p,L,U):
  from fc_linalg3D.simplelib import mtimes,eye
  (N,n,m)=A.shape
  K=np.tile(np.arange(N),(n,1)).T
  P=eye(N,n)[K,p]
  return np.max(abs(mtimes(L,U)-mtimes(P,A)))

def set_pALU(N,verbose,**kwargs):
  from fc_linalg3D.simplelib import randnsympd,ctranspose,mtimes
  Print=kwargs.get('Print',lambda s: print(s))
  d=kwargs.pop('d',[3,3]) 
  A,bdA=set_datas(N,verbose,Type='array3D',d=d,genfun=lambda N,m,n: randnsympd(N,m),**kwargs)
  error=lambda Out: error_pALU(A,*Out)
  return ((A,error),bdA)

def error_PALU(A,P,L,U):
  from fc_linalg3D import mtimes,eye
  return np.max(abs(mtimes(L,U)-mtimes(P,A)))


def set_PALU(N,verbose,**kwargs):
  from fc_linalg3D.random import randnsympd
  from fc_linalg3D import ctranspose,mtimes
  Print=kwargs.get('Print',lambda s: print(s))
  d=kwargs.pop('d',[3,3]) 
  A,bdA=set_datas(N,verbose,Type='array3D',d=d,genfun=lambda N,m,n: randnsympd(N,m),**kwargs)
  error=lambda Out: error_PALU(A,*Out)
  return ((A,error),bdA)

def error_pLUinplace(A,p1,A1):
  from fc_linalg3D import mtimes,eye
  from fc_linalg3D.fact import LUfrominplace,permInd2Mat
  #import copy
  #(N,n,m)=A.shape
  #K=np.tile(np.arange(N),(n,1)).T
  ##P1=eye(N,n)[K,p1]
  L1,U1=LUfrominplace(A1)
  P1=permInd2Mat(p1)
  return np.max(abs(mtimes(P1,A)-mtimes(L1,U1)))

def set_pLUinplace(N,verbose,**kwargs):
  from fc_linalg3D.random import randnsympd,randn
  Print=kwargs.get('Print',lambda s: print(s))
  d=kwargs.pop('d',2) 
  A,bdA=set_datas(N,verbose,Type='array3D',d=[d,d],genfun=lambda N,m,n: randn(N,m,n),**kwargs)
  Acopy=A.copy()
  error=lambda Out: error_pLUinplace(Acopy,*Out)
  return ((A,error),bdA)

  
def nameArray(Type,d):
  Type=Type.lower()
  if Type=='array3d':
    return Type+'-%dby%d'%(d[0],d[1])
  if Type=='array2d':
    return Type+'-%dby%d'%(d[0],d[1])
  if Type=='array1d':
    return Type
  if Type=='scalar':
    return Type
  return 'Unknow'

def print_bench(name,i,nb):
  print('\n================== %s [%2d/%2d] ==================\n'%(name,i,nb))

def bench_aXpbY(**kwargs):
  """ bench for aXpbY (linear combinations) functions
  
  from fc_linalg3D.benchs import bench_aXpbY
  bench_aXpbY(d=[3,4]) 
  bench_aXpbY(d=[4,3],slow=False,rtype='array2D')
  bench_aXpbY(d=[4,3],slow=False,ltype='array1D')
  """
  slow=kwargs.pop('slow',True)
  from fc_linalg3D.ebye import aXpbY_Cpt,aXpbY_Mat,aXpbY_Vec,aXpbY_Cvt,aXpbY_Nat
  if slow:
    c_default=10**2
    Lfun=[aXpbY_Cpt,aXpbY_Mat,aXpbY_Vec,aXpbY_Cvt,aXpbY_Nat]
  else:
    c_default=10**3
    Lfun=[aXpbY_Vec,aXpbY_Cvt,aXpbY_Nat]
  c=kwargs.pop('c',c_default)
  LN=kwargs.pop('LN',c*np.arange(2*10**2,10**3+1,2*10**2))
  Error=lambda out1,out2: np.max(abs(out1-out2))
  fc_bench.bench(Lfun,set_aXpbY,LN=LN,error=Error,**kwargs)

def bench_aXpbY_all(**kwargs):
  """ bench for aXpbY (linear combinations) functions
  
  from fc_linalg3D.benchs import bench_aXpbY
  bench_aXpbY_all(d=[3,4]) 
  bench_aXpbY_all(d=[4,3],slow=False) 
  """
  d=kwargs.get('d',[2,2]) 
  nb=2*len(TypeArray)-1 # number of benchs
  i=1
  name='bench_aXpbY'
  for rtype in TypeArray:
    print_bench(name,i,nb);i+=1
    tag=nameArray('array3D',d)+'-'+nameArray(rtype,d)
    bench_aXpbY(rtype=rtype,ltype='array3D',tag=tag,d=d,**kwargs)
  for ltype in TypeArray[1:]:
    print_bench(name,i,nb);i+=1
    tag=nameArray(ltype,d)+'-'+nameArray('array3D',d)
    bench_aXpbY(ltype=ltype,rtype='array3D',tag=tag,d=d,**kwargs)
    
def bench_mtimes(**kwargs):
  """ bench for mtimes (matricial product) functions
  
  from fc_linalg3D.benchs import bench_mtimes
  bench_mtimes(dl=[3,4],dr=[4,2]) 
  bench_mtimes(dl=[3,4],dr=[4,2],rtype='array2D',slow=False)
  """
  slow=kwargs.pop('slow',True)
  from fc_linalg3D.mtimes import mtimes_Cpt,mtimes_Mat,mtimes_Vec,mtimes_VecOpt,mtimes_Nat
  if slow:
    c_default=10**2
    Lfun=[mtimes_Cpt,mtimes_Mat,mtimes_Vec,mtimes_VecOpt,mtimes_Nat]
  else:
    c_default=10**3
    Lfun=[mtimes_Vec,mtimes_VecOpt,mtimes_Nat]
  c=kwargs.pop('c',c_default)
  LN=kwargs.pop('LN',c*np.arange(2*10**2,10**3+1,2*10**2))
  Error=lambda out1,out2: np.max(abs(out1-out2))
  fc_bench.bench(Lfun,set_mtimes,LN=LN,error=Error,**kwargs)
  
def bench_mtimes_all(**kwargs):
  """ bench for mtimes (matricial product) functions
  
  from fc_linalg3D.benchs import bench_mtimes_all
  bench_mtimes_all(dl=[3,4],dr=[4,2]) 
  bench_mtimes_all(dl=[3,4],dr=[4,2],slow=False)
  """
  dl=kwargs.pop('dl',[2,3])
  dr=kwargs.pop('dr',[3,2])
  tag=nameArray('array3D',dl)+'-'+nameArray('array3D',dr)
  bench_mtimes(ltype='array3D',dl=dl,rtype='array3D',dr=dr,tag=tag,**kwargs)
  tag=nameArray('array3D',dl)+'-'+nameArray('array2D',dr)
  bench_mtimes(ltype='array3D',dl=dl,rtype='array2D',dr=dr,tag=tag,**kwargs)
  tag=nameArray('array2D',dl)+'-'+nameArray('array3D',dr)
  bench_mtimes(ltype='array2D',dl=dl,rtype='array3D',dr=dr,tag=tag,**kwargs)
  
  
def bench_linsolve_diag(**kwargs):
  """ bench for linsolve_diag (solving diagonal systems) functions
  
  from fc_linalg3D.benchs import bench_linsolve_diag
  bench_linsolve_diag(d=4,m=2) 
  bench_linsolve_diag(d=4,m=2,rtype='array2D',slow=False) 
  """
  slow=kwargs.pop('slow',True)
  from fc_linalg3D.solver import linsolve_diag_Cpt,linsolve_diag_Mat,linsolve_diag_Vec,linsolve_Nat
  if slow:
    c_default=10**2
    Lfun=[linsolve_diag_Cpt,linsolve_diag_Mat,linsolve_diag_Vec,linsolve_Nat]
  else:
    c_default=10**3
    Lfun=[linsolve_diag_Vec,linsolve_Nat]
  c=kwargs.pop('c',c_default)
  LN=kwargs.pop('LN',c*np.arange(2*10**2,10**3+1,2*10**2))
  Error=lambda out1,out2: np.max(abs(out1-out2))
  fc_bench.bench(Lfun,set_linsolve_diag,LN=LN,error=Error,**kwargs)

def bench_linsolve_diag_all(**kwargs):
  """ bench for linsolve_diag (matrices multiplication) functions
  
  from fc_linalg3D.benchs import linsolve_diag_all
  bench_linsolve_diag_all(d=4,m=2) 
  bench_linsolve_diag_all(d=4,m=2,slow=False)  
  """
  d=kwargs.get('d',3)
  m=kwargs.get('m',2)
  dl=[d,d];dr=[d,m]
  tag=nameArray('array3D',dl)+'-'+nameArray('array3D',dr)
  bench_linsolve_diag(ltype='array3D',dl=dl,rtype='array3D',dr=dr,tag=tag,**kwargs)
  tag=nameArray('array3D',dl)+'-'+nameArray('array2D',dr)
  bench_linsolve_diag(ltype='array3D',dl=dl,rtype='array2D',dr=dr,tag=tag,**kwargs)
    
    
def bench_linsolve_tril(**kwargs):
  """ bench for linsolve_tril (solving lower triangular systems) functions
  
  from fc_linalg3D.benchs import bench_linsolve_tril
  bench_linsolve_tril(d=4,m=2) 
  bench_linsolve_tril(d=4,m=2,rtype='array2D',slow=False) 
  """
  slow=kwargs.pop('slow',True)
  from fc_linalg3D.solver import linsolve_tril_Cpt,linsolve_tril_Mat,linsolve_tril_Vec,linsolve_Nat
  if slow:
    c_default=10**2
    Lfun=[linsolve_tril_Cpt,linsolve_tril_Mat,linsolve_tril_Vec,linsolve_Nat]
  else:
    c_default=10**3
    Lfun=[linsolve_tril_Vec,linsolve_Nat]
  c=kwargs.pop('c',c_default)
  Comment='# Solving lower triangular systems by using fc_linalg3D.solver.linsolve_tril functions'
  LN=kwargs.pop('LN',c*np.arange(2*10**2,10**3+1,2*10**2))
  Error=lambda out1,out2: np.max(abs(out1-out2))
  fc_bench.bench(Lfun,set_linsolve_tril,LN=LN,error=Error,comment=Comment,**kwargs)
  
def bench_linsolve_tril_all(**kwargs):
  """ bench for linsolve_tril (solving lower triangular systems) functions
  
  from fc_linalg3D.benchs import bench_linsolve_tril_all
  bench_linsolve_tril_all(d=4,m=2) 
  bench_linsolve_tril_all(d=4,m=2,slow=False) 
  """
  d=kwargs.get('d',3)
  m=kwargs.get('m',2)
  dl=[d,d];dr=[d,m]
  tag=nameArray('array3D',dl)+'-'+nameArray('array3D',dr)
  bench_linsolve_tril(ltype='array3D',dl=dl,rtype='array3D',dr=dr,tag=tag,**kwargs)
  tag=nameArray('array3D',dl)+'-'+nameArray('array2D',dr)
  bench_linsolve_tril(ltype='array3D',dl=dl,rtype='array2D',dr=dr,tag=tag,**kwargs)

def bench_linsolve_triu(**kwargs):
  """ bench for linsolve_triu (solving upper triangular systems) functions
  
  from fc_linalg3D.benchs import bench_linsolve_triu
  bench_linsolve_triu(d=4,m=2) 
  bench_linsolve_triu(d=4,m=2,rtype='array2D',slow=False) 
  """
  slow=kwargs.pop('slow',True)
  from fc_linalg3D.solver import linsolve_triu_Cpt,linsolve_triu_Mat,linsolve_triu_Vec,linsolve_Nat
  if slow:
    c_default=10**2
    Lfun=[linsolve_triu_Cpt,linsolve_triu_Mat,linsolve_triu_Vec,linsolve_Nat]
  else:
    c_default=10**3
    Lfun=[linsolve_triu_Vec,linsolve_Nat]
  c=kwargs.pop('c',c_default)
  Comment='# Solving upper triangular systems by using fc_linalg3D.solver.linsolve_triu functions'
  LN=kwargs.pop('LN',c*np.arange(2*10**2,10**3+1,2*10**2))
  Error=lambda out1,out2: np.max(abs(out1-out2))
  fc_bench.bench(Lfun,set_linsolve_triu,LN=LN,error=Error,comment=Comment,**kwargs)
  
def bench_linsolve_triu_all(**kwargs):
  """ bench for linsolve_triu (solving upper triangular systems) functions
  
  from fc_linalg3D.benchs import bench_linsolve_triu_all
  bench_linsolve_triu_all(d=4,m=2) 
  bench_linsolve_triu_all(d=4,m=2,slow=False) 
  """
  d=kwargs.get('d',3)
  m=kwargs.get('m',2)
  dl=[d,d];dr=[d,m]
  tag=nameArray('array3D',dl)+'-'+nameArray('array3D',dr)
  bench_linsolve_triu(ltype='array3D',dl=dl,rtype='array3D',dr=dr,tag=tag,**kwargs)
  tag=nameArray('array3D',dl)+'-'+nameArray('array2D',dr)
  bench_linsolve_triu(ltype='array3D',dl=dl,rtype='array2D',dr=dr,tag=tag,**kwargs)
  
def bench_linsolve_PALU(**kwargs):
  """ bench for linsolve_PALU (solving regular linear systems) functions
  
  from fc_linalg3D.benchs import bench_linsolve_PALU
  bench_linsolve_PALU(d=4,m=2) 
  bench_linsolve_PALU(d=4,m=2,rtype='array2D') 
  """
  slow=kwargs.pop('slow',True)
  from fc_linalg3D.solver import linsolve_PALU_Vec,linsolve_Nat
  c_default=10**3
  Lfun=[linsolve_PALU_Vec,linsolve_Nat]
  c=kwargs.pop('c',c_default)
  LN=kwargs.pop('LN',c*np.arange(2*10**2,10**3+1,2*10**2))
  Error=lambda out1,out2: np.max(abs(out1-out2))
  Comment='# Solving systems by using fc_linalg3D.solver.linsolve_PALU function'
  fc_bench.bench(Lfun,set_linsolve_LU,LN=LN,error=Error,comment=Comment,**kwargs)  
    
def bench_linsolve_PALU_all(**kwargs):
  """ bench for linsolve_PALU (solving regular linear systems) functions
  
  from fc_linalg3D.benchs import bench_linsolve_PALU_all
  bench_linsolve_PALU_all(d=4,m=2) 
  """
  d=kwargs.get('d',3)
  m=kwargs.get('m',2)
  dl=[d,d];dr=[d,m]
  ltype='array3d'
  i=1
  for rtype in['array3d','array2d']:
    print('*'*50)
    print('* [%d/2] linsolve_PALU : ''%s'' vs ''%s'''%(i,ltype,rtype))
    print('*'*50)
    i+=1
    tag=nameArray(ltype,dl)+'-'+nameArray(rtype,dr)
    bench_linsolve_PALU(ltype=ltype,rtype=rtype,tag=tag,**kwargs)
  
def bench_linsolve_cholesky(**kwargs):
  """ bench for linsolve_cholesky functions
      Solve A X = B where A is a N-by-d-by-d 3D-array
      If B is N-by-d-by-n then X is N-by-d-by-n and
        A[k] @ X[k] = B[k]
      If B is d-by-n then X is N-by-d-by-n and
        A[k] @ X[k] = B  
        
  from fc_linalg3D.benchs import bench_linsolve_cholesky
  bench_linsolve_cholesky(d=3,n=2) 
  bench_linsolve_cholesky(d=4,n=1,rtype='array2D') 
  """
  c=kwargs.pop('c',10**3)
  LN=kwargs.pop('LN',c*np.arange(10**2,10**3+1,2*10**2))
  d=kwargs.pop('d',2) # in [2,2.5,3]
  m=kwargs.pop('m',1) # in [2,2.5,3]
  from fc_linalg3D.solver import linsolve_cholesky,linsolve_Nat
  Lfun=[linsolve_cholesky,linsolve_Nat]
  Error=lambda out1,out2: np.max(abs(out1-out2))
  Comment='# Solving systems by using fc_linalg3D.solver.linsolve_cholesky function'
  fc_bench.bench(Lfun,set_linsolve_cholesky,LN=LN,error=Error,comment=Comment,**kwargs)  
  
def bench_linsolve_cholesky_all(**kwargs):
  """ bench for linsolve_cholesky functions
      Solve A X = B where A is a N-by-d-by-d 3D-array
      If B is N-by-d-by-n then X is N-by-d-by-n and
        A[k] @ X[k] = B[k]
      If B is d-by-n then X is N-by-d-by-n and
        A[k] @ X[k] = B  
        
  from fc_linalg3D.benchs import bench_linsolve_cholesky_all
  bench_linsolve_cholesky_all(d=3,n=2) 
  bench_linsolve_cholesky_all(d=4,n=1) 
  """
  d=kwargs.pop('d',2)
  m=kwargs.pop('m',1)
  dl=[d,d];dr=[d,m]
  ltype='array3d'
  i=1
  for rtype in['array3d','array2d']:
    print('*'*50)
    print('* [%d/2] linsolve_cholesky : ''%s'' vs ''%s'''%(i,ltype,rtype))
    print('*'*50)
    i+=1
    tag=nameArray(ltype,dl)+'-'+nameArray(rtype,dr)
    bench_linsolve_cholesky(ltype=ltype,rtype=rtype,tag=tag,**kwargs)
  
  
def bench_binaryop(**kwargs):
  """ bench for binaryop (element by element binary operator) functions
  
  from fc_linalg3D.benchs import bench_binaryop
  bench_binaryop(d=[3,4]) 
  bench_binaryop(d=[3,4],rtype='array2d') 
  """
  op=kwargs.get('op','add') # in ['add','sub','times',...]
  slow=kwargs.pop('slow',True)
  import fc_linalg3D.ebye as ebye
  if slow:
    c_default=10**2
    Lfun=[eval('ebye.'+op+'_Cpt'),eval('ebye.'+op+'_Mat'),eval('ebye.'+op+'_Vec'),eval('ebye.'+op+'_Cvt')]
    names=[op+'_Cpt',op+'_Mat',op+'_Vec',op+'_Cvt']
  else:
    c_default=10**3
    Lfun=[eval('ebye.'+op+'_Vec'),eval('ebye.'+op+'_Cvt')]
    names=[op+'_Vec',op+'_Cvt']
  c=kwargs.pop('c',c_default)
  LN=kwargs.pop('LN',c*np.arange(2*10**2,10**3+1,2*10**2))
  Error=lambda out1,out2: np.max(abs(out1-out2))
  Comment='# Computing %s binary operator function: %s(A,B) function'%(op,op)
  fc_bench.bench(Lfun,set_ebye,LN=LN,names=names,error=Error,comment=Comment,**kwargs)
 
def bench_binaryop_all(**kwargs):
  """ bench for binaryop (element by element binary operator) functions
  
  from fc_linalg3D.benchs import bench_binaryop_all
  bench_binaryop_all(d=[3,4]) 
  bench_binaryop_all(d=[3,4],op='times') 
  """
  d=kwargs.pop('d',[2,3])
  op=kwargs.pop('op','add')
  i=1;nb=7;
  ltype='array3D'
  for rtype in TypeArray:
    print('*'*50)
    print('* [%d/%d] binaryop<%s> : ''%s'' vs ''%s'''%(i,nb,op,ltype,rtype))
    print('*'*50)
    i+=1
    tag=nameArray('array3D',d)+'-'+nameArray(rtype,d)
    bench_binaryop(op=op,rtype=rtype,ltype='array3D',tag=tag,**kwargs)
  rtype='array3D'
  for ltype in TypeArray[1:]:
    print('*'*50)
    print('* [%d/%d] binaryop<%s> : ''%s'' vs ''%s'''%(i,nb,op,ltype,rtype))
    print('*'*50)
    i+=1
    tag=nameArray(ltype,d)+'-'+nameArray('array3D',d)
    bench_binaryop(op=op,rtype='array3D',ltype=ltype,tag=tag,d=d,**kwargs)
    
def bench_plus_all(**kwargs):
  bench_binaryop_all(op='plus',**kwargs)

def bench_minus_all(**kwargs):
  bench_binaryop_all(op='minus',**kwargs)

def bench_times_all(**kwargs):
  bench_binaryop_all(op='times',**kwargs)
  
def bench_rdivide_all(**kwargs):
  bench_binaryop_all(op='rdivide',**kwargs)
  
def bench_ldivide_all(**kwargs):
  bench_binaryop_all(op='ldivide',**kwargs)  

def bench_PALU(**kwargs):
  """ bench for all PALU_* functions.
      Computing permuted LU factorization
  
  from fc_linalg3D.benchs import bench_PALU
  bench_PALU(d=[4,4]) 
  """
  d=kwargs.pop('d',[3,3])
  slow=kwargs.pop('slow',True)
  from fc_linalg3D.fact import PALU_Cpt,PALU_Mat,PALU_Vec
  if slow:
    c_default=10**2
    Lfun=[PALU_Cpt,PALU_Mat,PALU_Vec]
  else:
    c_default=10**3
    Lfun=[PALU_Vec]
  c=kwargs.pop('c',c_default)
  LN=kwargs.pop('LN',c*np.arange(2*10**2,10**3+1,2*10**2))
  Comment='# Computing LU factorization with partial pivoting by using fc_linalg3D.fact.PALU functions'
  Error=lambda out1,out2: np.max(abs(out1[0]-out2[0]))+np.max(abs(out1[1]-out2[1]))+np.max(abs(out1[2]-out2[2]))
  fc_bench.bench(Lfun,set_PALU,d=d,LN=LN,error=Error,**kwargs)
  

def bench_pLUinplace(**kwargs):
  """ bench for pLUinplace functions
  
  from fc_linalg3D.benchs import bench_pLUinplace
  bench_pLUinplace(d=[4,4]) 
  """
  d=kwargs.pop('d',[3,3])
  slow=kwargs.pop('slow',True)
  from fc_linalg3D.fact import pLUinplace_Cpt,pLUinplace_Mat,pLUinplace_Vec,pLUinplace_Vec01
  if slow:
    c_default=10**2
    Lfun=[pLUinplace_Cpt,pLUinplace_Mat,pLUinplace_Vec,pLUinplace_Vec01]
  else:
    c_default=10**3
    Lfun=[pLUinplace_Vec,pLUinplace_Vec01]
  c=kwargs.pop('c',c_default)
  LN=kwargs.pop('LN',c*np.arange(2*10**2,10**3+1,2*10**2))
  Err=lambda out1,out2: np.max(abs(out1[0]-out2[0]))+np.max(abs(out1[1]-out2[1]))
  fc_bench.bench(Lfun,set_datas,d=d,LN=LN,copyinput=True,error=Err,**kwargs)#,error=Err,**kwargs)
  
def bench_cholesky(**kwargs):
  """ bench for cholesky functions
  
  from fc_linalg3D.benchs import bench_cholesky
  bench_cholesky(d=4) 
  bench_cholesky(d=4,slow=False)
  bench_cholesky(d=4,slow=False,nbruns=1)
  """
  d=kwargs.pop('d',3)
  slow=kwargs.pop('slow',True)
  from fc_linalg3D.fact import cholesky_Cpt,cholesky_Mat,cholesky_Mat01,cholesky_Vec,cholesky_Vec01,cholesky_Nat
  if slow:
    c_default=10**2
    Lfun=[cholesky_Cpt,cholesky_Mat,cholesky_Mat01,cholesky_Vec,cholesky_Vec01,cholesky_Nat]
  else:
    c_default=10**3
    Lfun=[cholesky_Vec,cholesky_Vec01,cholesky_Nat]
  c=kwargs.pop('c',c_default)
  LN=kwargs.pop('LN',c*np.arange(2*10**2,10**3+1,2*10**2))
  Error=lambda out1,out2: np.max(abs(out1-out2))
  fc_bench.bench(Lfun,set_cholesky,d=d,LN=LN,error=Error,**kwargs)

  
def bench_det(**kwargs):
  """ bench for cholesky functions
  
  from fc_linalg3D.benchs import bench_cholesky
  bench_det(d=4) 
  bench_det(d=4,slow=False)
  bench_det(d=4,slow=False,nbruns=1)
  """
  d=kwargs.pop('d',3)
  slow=kwargs.pop('slow',True)
  from fc_linalg3D.det import Laplace_Mat,Laplace_Cpt,Laplace_Vec,LapIdx_Vec,detMix_Vec,det_pLUinplace,det_Nat
  if slow:
    c_default=1
    Lfun=[Laplace_Mat,Laplace_Cpt,Laplace_Vec,LapIdx_Vec,detMix_Vec,det_pLUinplace,det_Nat]
  else:
    c_default=10**3
    Lfun=[Laplace_Vec,LapIdx_Vec,detMix_Vec,det_pLUinplace,det_Nat]
  c=kwargs.pop('c',c_default)
  LN=kwargs.pop('LN',c*np.arange(2*10**2,10**3+1,2*10**2))
  Error=lambda out1,out2: np.max(abs(out1-out2))
  fc_bench.bench(Lfun,set_sympd,d=d,LN=LN,error=Error,**kwargs)

