from __future__ import print_function
import numpy as np
from fc_bench.bench import line_text_delimiter

def all_demos(N=10**4):
  Lfun=[demo01,demo02,demo03,demo04]
  ltd=line_text_delimiter()
  for fun in Lfun:
    print(ltd)
    print('  Running fc_linalg3D.demos.%s function'%fun.__name__)
    print(ltd)
    fun(N)

def demo01(N=10**4):
  import fc_linalg3D
  A=fc_linalg3D.random.randnsympd(N,3) # Needed with detMix_Vec and detMixIdx_Vec
  D=fc_linalg3D.det.LapIdx_Vec(A) # or any of the fc_linalg3D.det.det* functions
  D1=np.zeros(D.shape)
  for i in np.arange(N):
    D1[i]=np.linalg.det(A[i])
  print('Determinants error: %.4e'%np.linalg.norm(D-D1, np.inf))

def demo02(N=10**4):
  import fc_linalg3D
  d=np.random.randint(2,5)
  print('Generating A a N-by-d-by-d 3D-array where ');
  print('    A(k,:,:) are symmetric positive definite matrices')
  print('and N=%d, d=%d.'%(N,d))
  A=fc_linalg3D.random.randnsympd(N,d)
  b=np.ones((d,1))
  print('Solving  A*X=b where b is the ones vector')
  X=fc_linalg3D.solver.linsolve_cholesky_Vec(A,b)
  b1=fc_linalg3D.mtimes(A,X) # b1 is a (N,d) array
  print('Computes b1=A*X (must be the ones vector)')
  print(b1)
  print('Computes E=|b1-b| (must be the ones vector)\n')
  E=np.abs(b1-b); # Unsing broadcating. Computes |A*X-b|
  print(E);
  print('Error: %.4e'%np.max(E))
  
def demo03(N=10**4):
  import fc_linalg3D
  d=np.random.randint(2,5)
  print('Generating A a N-by-d-by-d 3D-array where ');
  print('    A(k,:,:) are symmetric positive definite matrices')
  print('and N=%d, d=%d.'%(N,d))
  A=fc_linalg3D.random.randnsympd(N,d)
  B=fc_linalg3D.random.randn(N,d,2)
  print('Solving  A*X=B where B is a random N-by-d-by-2 3D-array')
  X=fc_linalg3D.solver.linsolve_cholesky_Vec(A,B)
  print('Computes B1=A*X')
  B1=fc_linalg3D.mtimes(A,X) # B1 is a (N,d,2) array
  print('Computes E=|B1-B| (must be the ones vector)')
  E=np.abs(B1-B) # Computes |A*X-B|
  print(E)
  print('Error: %.4e'%np.max(E))
  
def demo04(N=10**4):
  import fc_linalg3D
  d=np.random.randint(2,5)
  print('Generating A a %d-by-%d-by-%d 3D-array with '%(N,d,d));
  print('    A(k,:,:)=L(k,:,:)*U(k,:,:), for all k')
  print('where L(k,:,:) and U(k,:,:) are respectively a lower and an upper triangular matrices')
  L=fc_linalg3D.random.randtril(N,d);
  U=fc_linalg3D.random.randtriu(N,d)
  A=fc_linalg3D.mtimes(L,U)
  
  Dex=fc_linalg3D.simplelib.diagprod(L)*fc_linalg3D.simplelib.diagprod(U)
  
  funs=[fc_linalg3D.det.det_Nat , fc_linalg3D.det.Laplace_Cpt,fc_linalg3D.det.Laplace_Mat, 
        fc_linalg3D.det.Laplace_Vec, fc_linalg3D.det.LapIdx_Vec, 
        fc_linalg3D.det.det_pLUinplace_Vec]
  
  for fun in funs:
    D=fun(A)
    print('Error=%.4f with %s function'%(np.max(np.abs(D-Dex)),fun.__name__))
  

  
  
