import numpy as np

def det_Nat(A):
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert m==n
  return np.linalg.det(A)
  
def Laplace(A):
  n=A.shape[0]
  if n==1:
    return A[0,0]
  det=0
  for j in np.arange(n):
    det+=(-1)**(j)*A[0,j]*Laplace(A[1:,np.setdiff1d(np.arange(n),j)])
  return det

def Laplace_Mat(A):
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert m==n
  if n==1:
    return A[:,0,0]
  D=np.zeros((N,))
  for k in np.arange(N):
    D[k]=Laplace(A[k])
  return D

def Laplace_Cpt(A):
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert m==n
  if n==1:
    return A[:,0,0]
  D=np.zeros((N,))
  for k in np.arange(N):
    for j in np.arange(n):
      D[k]+=(-1)**(j)*A[k,0,j]*Laplace(A[k,1:,np.setdiff1d(np.arange(n),j)])
  return D

def Laplace_Vec(A):
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert m==n
  if n==1:
    return A[:,0,0]
  D=np.zeros((N,))
  for j in np.arange(n):
    D+=(-1)**(j)*A[:,0,j]*Laplace_Vec(A[:,1:,np.setdiff1d(np.arange(n),j)])
  return D

def LapIdx_Vec(A,I=None,J=None):
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert m==n
  if I is None:
    I=np.arange(m)
  if J is None:
    J=np.arange(m)
  d=J.shape[0]
  if d==1:
    return A[:,I[0],J[0]]
  if d==2:
    return A[:,I[0],J[0]]*A[:,I[1],J[1]]-A[:,I[0],J[1]]*A[:,I[1],J[0]]
  
  D=np.zeros((N,))
  for j in np.arange(d):
    D+=(-1)**(j)*A[:,I[0],J[j]]*LapIdx_Vec(A,I[1:],J[np.setdiff1d(np.arange(d),j)])
  return D

def detMix_Vec(A):
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert m==n
  if n==1:
    return A[:,0,0]
  if n==2:
    return A[:,0,0]*A[:,1,1]-A[:,0,1]*A[:,1,0]
  A11=A[:,1:,1:]
  A1d=A[:,1:,:-1]
  Add=A[:,:-1,:-1]
  A11dd=A[:,1:-1,1:-1]
  return ( detMix_Vec(A11)*detMix_Vec(Add)- Laplace_Vec(A1d)**2 )/detMix_Vec(A11dd)

def detMixIdx_Vec(A,I=None,J=None):
  from fc_linalg3D.simplelib import is3Darray
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert m==n
  if I is None:
    I=np.arange(m)
  if J is None:
    J=np.arange(m)
  d=J.shape[0]
  if d==1:
    return A[:,I[0],J[0]]
  if d==2:
    return A[:,I[0],J[0]]*A[:,I[1],J[1]]-A[:,I[0],J[1]]*A[:,I[1],J[0]]
  I1=I[1:];J1=J[1:]
  Id=I[:-1];Jd=J[:-1]
  I2=I[1:-1];J2=J[1:-1]
  return ( detMixIdx_Vec(A,I1,J1)*detMixIdx_Vec(A,Id,Jd)- LapIdx_Vec(A,I1,Jd)**2 )/detMixIdx_Vec(A,I2,J2)

def det_pLUinplace_Vec(A):
  from fc_linalg3D.fact import pLUinplace_Vec,permsign
  from fc_linalg3D.simplelib import diagprod
  Atmp=A.copy()
  p,Atmp=pLUinplace_Vec(Atmp)
  return permsign(p)*diagprod(Atmp)

def det_pLUinplace(A):
  return det_pLUinplace_Vec(A)

  
def det_Nat(A):
  return np.linalg.det(A)
