import numpy as np
from fc_linalg3D.simplelib import is3Darray,getCpts,getMats,getVecs,getNdims,to3Darray

def mtimes_Nat(X,Y):
  return np.matmul(X,Y)

def mtimes_Cpt(X,Y):
  """  returns X*Y
  """
  isX3D=is3Darray(X)
  isY3D=is3Darray(Y)
  assert isX3D or isY3D, 'One of the parameters must be an 3D-array'
  getX,getY,N,dimX,dimY=getCpts(X,Y)
  assert (dimX[2]==dimY[1])
  C=np.zeros((N,dimX[1],dimY[2]))
  for k in range(N):
    for i in np.arange(dimX[1]):
      for j in np.arange(dimY[2]):
        S=0
        for l in np.arange(dimX[2]):
          S+=getX(X,k,i,l)*getY(Y,k,l,j)
        C[k,i,j]=S
  return C   

def mtimes_Mat(X,Y):
  """  returns X*Y
  """
  isX3D=is3Darray(X)
  isY3D=is3Darray(Y)
  assert isX3D or isY3D, 'One of the parameters must be an 3D-array'
  getX,getY,N,dimX,dimY=getMats(X,Y)
  assert dimX[2]==dimY[1]
  C=np.zeros((N,dimX[1],dimY[2]))
  for k in range(N):
    C[k]=np.matmul(getX(X,k),getY(Y,k))
  return C     
 
def mtimes_Vec(X,Y):
  """  returns X*Y
  """
  isX3D=is3Darray(X)
  isY3D=is3Darray(Y)
  assert isX3D or isY3D, 'One of the parameters must be an 3D-array'
  getX,getY,N,dimX,dimY=getVecs(X,Y)
  assert dimX[2]==dimY[1]
  C=np.zeros((N,dimX[1],dimY[2]))
  S=np.zeros((N,))
  for i in np.arange(dimX[1]):
    for j in np.arange(dimY[2]):
      S=0*S
      for l in np.arange(dimX[2]):
        S+=getX(X,i,l)*getY(Y,l,j)
      C[:,i,j]=S
  return C

def mtimes_Cvt(X,Y):
  """  returns X*Y
  """
  isX3D=is3Darray(X)
  isY3D=is3Darray(Y)
  assert isX3D or isY3D, 'One of the parameters must be an 3D-array'
  getX,getY,N,dimX,dimY=getVecs(X,Y)
  assert dimX[2]==dimY[1]
  X3D=to3Darray(X,N,dimX[1],dimX[2])
  Y3D=to3Darray(Y,N,dimY[1],dimY[2])
  return np.matmul(X3D,Y3D)

def mtimes_VecOpt(X,Y):
  """  returns X*Y
  """
  isX3D=is3Darray(X)
  isY3D=is3Darray(Y)
  assert isX3D or isY3D, 'One of the parameters must be an 3D-array'
  getX,getY,N,dimX,dimY=getVecs(X,Y)
  assert dimX[2]==dimY[1]
  C=np.zeros((N,dimX[1],dimY[2]))
  S=np.zeros((N,))
  for i in np.arange(dimX[1]):
    for j in np.arange(dimY[2]):
      S=0*S
      for l in np.arange(dimX[2]):
        S+=getX(X,i,l)*getY(Y,l,j)
      C[:,i,j]=S
  return C 
 
def mtimes(X,Y):
  return np.matmul(X,Y)
