import numpy as np

def randn(N,m,n):
  """ random N m-by-n  matrices
  """
  return np.random.randn(N,m,n)
    
def randnsym(N,d):
  """ random N d-by-d symmetric matrices
  """
  A=np.zeros((N,d,d))
  for i in np.arange(d):
    A[:,i,i]=np.random.randn(N)
    for j in np.arange(i+1,d):
      A[:,i,j]=np.random.randn(N)
      A[:,j,i]=np.random.randn(N)
  return A

def randnsdd(N,d):
  """ random N d-by-d strict diagonally dominant matrices
  """
  A=randn(N,d,d)
  T=abs(A).sum(2)
  for i in np.arange(d):
    A[:,i,i]=T[:,i]
  return A

def randnsympd(N,d):
  """ random N d-by-d symmetric positive definite matrices
  """
  from fc_linalg3D import mtimes,transpose
  A=randnsdd(N,d) # generate N regular d-by-d matrices (as stric diagonally dominante matrices)
  return mtimes(A,transpose(A))

def randn_diag(N,d,**kwargs):
  """ random N d-by-d diagonal matrices
  """
  m=kwargs.pop('m',0)
  sigma=kwargs.pop('sigma',1)
  X=np.zeros((N,d,d))
  for i in np.arange(d):
    X[:,i,i]=m+sigma*np.random.randn(N,**kwargs)
  return X

def randdiag(N,d,**kwargs):
  """ random N d-by-d diagonal matrices
  """
  a=kwargs.pop('a',0)
  b=kwargs.pop('b',1)
  X=np.zeros((N,d,d))
  for i in np.arange(d):
    X[:,i,i]=a+(b-a)*np.random.rand(N,**kwargs)
  return X

def randintdiag(N,d,**kwargs):
  """ Return random integers from the 'discrete uniform' distribution of the
      specified dtype in the 'half-open' interval [low, high). 
      If high is None (the default), then results are from [0, low).
      By default low is 10.
  """
  a=kwargs.pop('a', 0)
  b=kwargs.pop('b', 10)
  X=np.zeros((N,d,d))
  for i in np.arange(d):
    X[:,i,i]=np.random.randint(low=a,high=b,size=(N),**kwargs)
  return X

def randntriu(N,d,**kwargs):
  """ random N d-by-d triangular upper matrices
  """
  m=kwargs.pop('m',0)
  sigma=kwargs.pop('sigma',1)
  X=np.zeros((N,d,d))
  for i in np.arange(d):
    for j in np.arange(i,d):
      X[:,i,j]=m+sigma*np.random.randn(N,**kwargs)
  return X

def randtriu(N,d,**kwargs):
  """ random N d-by-d diagonal matrices
  """
  a=kwargs.pop('a',0)
  b=kwargs.pop('b',1)
  X=np.zeros((N,d,d))
  for i in np.arange(d):
    for j in np.arange(i,d):
      X[:,i,j]=a+(b-a)*np.random.rand(N,**kwargs)
  return X

def randinttriu(N,d,**kwargs):
  """
  Return random integers from the 'discrete uniform' distribution of the 
  specified dtype in the 'half-open' interval [low, high). 
  If high is None (the default), then results are from [0, low).
  By default low is 10.
  """
  la=kwargs.pop('a', 0)
  b=kwargs.pop('b', 10)
  X=np.zeros((N,d,d))
  for i in np.arange(d):
    for j in np.arange(i,d):
      X[:,i,j]=np.random.randint(low=a,high=b,size=(N),**kwargs)
  return X

def randntril(N,d,**kwargs):
  """ random N d-by-d triangular upper matrices
  """
  m=kwargs.pop('m',0)
  sigma=kwargs.pop('sigma',1)
  X=np.zeros((N,d,d))
  for i in np.arange(d):
    for j in np.arange(0,i+1):
      X[:,i,j]=m+sigma*np.random.randn(N,**kwargs)
  return X

def randtril(N,d,**kwargs):
  """ random N d-by-d diagonal matrices
  """
  a=kwargs.pop('a',0)
  b=kwargs.pop('b',1)
  X=np.zeros((N,d,d))
  for i in np.arange(d):
    for j in np.arange(0,i+1):
      X[:,i,j]=a+(b-a)*np.random.rand(N,**kwargs)
  return X

def randinttril(N,d,**kwargs):
  """
  Return random integers from the 'discrete uniform' distribution of the 
  specified dtype in the 'half-open' interval [low, high). 
  If high is None (the default), then results are from [0, low).
  By default low is 10.
  """
  a=kwargs.pop('a', 0)
  b=kwargs.pop('b', 10)
  X=np.zeros((N,d,d))
  for i in np.arange(d):
    for j in np.arange(0,i+1):
      X[:,i,j]=np.random.randint(low=a,high=b,size=(N),**kwargs)
  return X
