import numpy as np

from fc_linalg3D.simplelib import is3Darray,getScaComp,getComp,ctranspose,zeros,eye,getCpt

""" Contains solvers for 3D-array
         A X = B 
    where A is a N-by-d-by-d 3D-array
    - If B is N-by-d-by-n then X is the N-by-d-by-n 3D-array such that
        A[k] @ X[k] = B[k]
    - If B is d-by-n then X is the N-by-d-by-n 3D-array such that
        A[k] @ X[k] = B  
"""

def linsolve_Nat(A,b):
  bb=b
  if not is3Darray(b):
    bb=b.reshape((1,b.shape[0],b.shape[1]))
  return np.linalg.solve(A,bb)

def linsolve_diag_v00(A,b):
  assert is3Darray(A) 
  (N,m,n)=A.shape
  (getb,(Nb,mb,nb))=getScaComp(b,N)
  assert n==mb
  X=np.zeros((N,m,nb))
  if nb==1:
    getij=lambda b,i,j: b[i]
  else:
    getij=lambda b,i,j: b[i,j]

  for k in np.arange(N):
    bk=getb(b,k)
    for i in np.arange(m):
      for j in np.arange(nb):
        X[k,i,j]=getij(bk,i,j)/A[k,i,i]
  return X

def linsolve_diag_v01(A,b):
  assert is3Darray(A) 
  (N,m,n)=A.shape
  (getb,(Nb,mb,nb))=getComp(b,N)
  assert n==mb
  X=np.zeros((N,m,nb))
  for i in np.arange(m):
    for j in np.arange(nb):
      X[:,i,j]=getb(b,i,j)/A[:,i,i]
  return X

def matricial_linsolve_diag(A,B):
  (d,p)=A.shape
  assert d==p and B.shape[0]==d
  m=B.shape[1]
  X=np.zeros((d,m))
  for j in np.arange(m):
    for i in np.arange(d):
      X[i,j]=B[i,j]/A[i,i]
  return X

def linsolve_diag_Cpt(A,B):
  from fc_linalg3D.simplelib import getCpt
  assert is3Darray(A) 
  (N,d,p)=A.shape
  assert d==p
  getB,dimB=getCpt(B,N)
  assert dimB[1]==d
  m=dimB[2]
  X=np.zeros((N,d,m))
  for k in np.arange(N): 
    for j in np.arange(m):
      for i in np.arange(d):
        X[k,i,j]=getB(B,k,i,j)/A[k,i,i]
  return X

def linsolve_diag_Mat(A,B):
  from fc_linalg3D.simplelib import getMat
  assert is3Darray(A) 
  (N,d,p)=A.shape
  assert d==p
  getB,dimB=getMat(B,N)
  assert dimB[1]==d
  m=dimB[2]
  X=np.zeros((N,d,m))
  for k in np.arange(N): 
    X[k]=matricial_linsolve_diag(A[k],getB(B,k))
  return X

def linsolve_diag_Vec(A,B):
  from fc_linalg3D.simplelib import getVec
  assert is3Darray(A) 
  (N,d,p)=A.shape
  assert d==p
  getB,dimB=getVec(B,N)
  assert dimB[1]==d
  m=dimB[2]
  X=np.zeros((N,d,m))
  for j in np.arange(m):
    for i in np.arange(d):
      X[:,i,j]=getB(B,i,j)/A[:,i,i]
  return X

def matricial_linsolve_tril(A,B):
  (d,p)=A.shape
  assert d==p and B.shape[0]==d
  m=B.shape[1]
  X=np.zeros((d,m))
  for l in np.arange(m):
    for i in np.arange(d):
      S=0
      for j in np.arange(i):
        S+=A[i,j]*X[j,l]
      X[i,l]=(B[i,l]-S)/A[i,i]
  return X

def linsolve_tril_Mat(A,B):
  from fc_linalg3D.simplelib import getMat
  assert is3Darray(A) 
  (N,d,p)=A.shape
  assert d==p
  getB,dimB=getMat(B,N)
  assert dimB[1]==d
  m=dimB[2]
  X=np.zeros((N,d,m))
  for k in np.arange(N): 
    X[k]=matricial_linsolve_tril(A[k],getB(B,k))
  return X

def linsolve_tril_Cpt(A,B):
  from fc_linalg3D.simplelib import getCpt
  assert is3Darray(A) 
  (N,d,p)=A.shape
  assert d==p
  getB,dimB=getCpt(B,N)
  assert dimB[1]==d
  m=dimB[2]
  X=np.zeros((N,d,m))
  for k in np.arange(N): 
    for l in np.arange(m):
      for i in np.arange(d):
        S=0
        for j in np.arange(i):
          S+=A[k,i,j]*X[k,j,l]
        X[k,i,l]=(getB(B,k,i,l)-S)/A[k,i,i]
  return X

def linsolve_tril_Vec(A,B):
  from fc_linalg3D.simplelib import getVec
  assert is3Darray(A) 
  (N,d,p)=A.shape
  assert d==p
  getB,dimB=getVec(B,N)
  assert dimB[1]==d
  m=dimB[2]
  X=np.zeros((N,d,m))
  S=np.zeros((N,))
  for l in np.arange(m):
    for i in np.arange(d):
      S=0*S
      for j in np.arange(i):
        S+=A[:,i,j]*X[:,j,l]
      X[:,i,l]=(getB(B,i,l)-S)/A[:,i,i]
  return X

def matricial_linsolve_triu(A,B):
  (d,p)=A.shape
  assert d==p and B.shape[0]==d
  m=B.shape[1]
  X=np.zeros((d,m))
  for l in np.arange(m):
    for i in np.arange(d-1,-1,-1):
      S=0
      for j in np.arange(i+1,d):
        S+=A[i,j]*X[j,l]
      X[i,l]=(B[i,l]-S)/A[i,i]
  return X

def linsolve_triu_Mat(A,B):
  from fc_linalg3D.simplelib import getMat
  assert is3Darray(A) 
  (N,d,p)=A.shape
  assert d==p
  getB,dimB=getMat(B,N)
  assert dimB[1]==d
  m=dimB[2]
  X=np.zeros((N,d,m))
  for k in np.arange(N): 
    X[k]=matricial_linsolve_triu(A[k],getB(B,k))
  return X

def linsolve_triu_Cpt(A,B):
  from fc_linalg3D.simplelib import getCpt
  assert is3Darray(A) 
  (N,d,p)=A.shape
  assert d==p
  getB,dimB=getCpt(B,N)
  assert dimB[1]==d
  m=dimB[2]
  X=np.zeros((N,d,m))
  for k in np.arange(N): 
    for l in np.arange(m):
      for i in np.arange(d-1,-1,-1):
        S=0
        for j in np.arange(i+1,d):
          S+=A[k,i,j]*X[k,j,l]
        X[k,i,l]=(getB(B,k,i,l)-S)/A[k,i,i]
  return X

def linsolve_triu_Vec(A,B):
  from fc_linalg3D.simplelib import getVec
  assert is3Darray(A) 
  (N,d,p)=A.shape
  assert d==p
  getB,dimB=getVec(B,N)
  assert dimB[1]==d
  m=dimB[2]
  X=np.zeros((N,d,m))
  S=np.zeros((N,))
  for l in np.arange(m):
    for i in np.arange(d-1,-1,-1):
      S=0*S
      for j in np.arange(i+1,d):
        S+=A[:,i,j]*X[:,j,l]
      X[:,i,l]=(getB(B,i,l)-S)/A[:,i,i]
  return X

def linsolve_PALU_Cpt(A,b):
  """ A is a regular 3D-array 
  """
  from fc_linalg3D.fact import PALU_Cpt
  from fc_linalg3D.mtimes import mtimes_Cpt
  [P,L,U]=PALU_Cpt(A)
  Y=linsolve_tril_Cpt(L,mtimes_Cpt(P,b))
  return linsolve_triu_Cpt(U,Y)

def linsolve_PALU_Mat(A,b):
  """ A is a regular 3D-array 
  """
  from fc_linalg3D.fact import PALU_Mat
  from fc_linalg3D.mtimes import mtimes_Mat
  [P,L,U]=PALU_Mat(A)
  Y=linsolve_tril_Mat(L,mtimes_Mat(P,b))
  return linsolve_triu_Mat(U,Y)

def linsolve_PALU_Vec(A,b):
  """ A is a regular 3D-array 
  """
  from fc_linalg3D.fact import PALU_Vec
  from fc_linalg3D.mtimes import mtimes_Vec
  [P,L,U]=PALU_Vec(A)
  Y=linsolve_tril_Vec(L,mtimes_Vec(P,b))
  return linsolve_triu_Vec(U,Y)

def linsolve_cholesky_Cpt(A,b):
  """ A is a symmetric positive definite 3D-array 
  """
  from fc_linalg3D.fact import cholesky_Cpt
  from fc_linalg3D import ctranspose
  L=cholesky_Cpt(A)
  Y=linsolve_tril_Cpt(L,b)
  return linsolve_triu_Cpt(ctranspose(L),Y)

def linsolve_cholesky_Mat(A,b):
  """ A is a symmetric positive definite 3D-array 
  """
  from fc_linalg3D.fact import cholesky_Mat
  from fc_linalg3D import ctranspose
  L=cholesky_Mat(A)
  Y=linsolve_tril_Mat(L,b)
  return linsolve_triu_Mat(ctranspose(L),Y)

def linsolve_Nat(A,b):
  """ A is a regular 3D-array 
  """
  if is3Darray(b):
    return np.linalg.solve(A,b)
  if b.shape==A.shape[:-1]:
    return np.linalg.solve(A,b)
  if b.shape[0]!=A.shape[0] and b.shape[0]==A.shape[1]:
    return np.linalg.solve(A,np.reshape(b,[1,b.shape[0],b.shape[1]]))
  raise ValueError('Trouble with dimension A:%s, b=%s'%(str(A.shape),str(b.shape)))

def linsolve_cholesky_Vec(A,b):
  """ A is a symmetric positive definite 3D-array 
  """
  from fc_linalg3D.fact import cholesky_Vec
  from fc_linalg3D.simplelib import ctranspose
  L=cholesky_Vec(A)
  Y=linsolve_tril_Vec(L,b)
  return linsolve_triu_Vec(ctranspose(L),Y)

def linsolve_cholesky_Nat(A,b):
  """ A is a symmetric positive definite 3D-array 
      Only use numpy cholesky factorization
      No broadcasting available in scipy.linalg for using solve_triangular
  """
  from fc_linalg3D.fact import cholesky_Nat
  from fc_linalg3D.simplelib import ctranspose
  L=cholesky_Nat(A)
  Y=linsolve_tril_Vec(L,b)
  return linsolve_triu_Vec(ctranspose(L),Y)

def linsolve_cholesky(A,b):
    return linsolve_cholesky_Nat(A,b)

def linsolve_triu_sca(A,b):
  (m,n)=A.shape
  (mb,nb)=b.shape
  assert m==n and mb==m
  X=np.zeros((m,nb))
  for l in np.arange(nb): # Solve A X[:,l]=b[:,l]
    for i in np.arange(m-1,-1,-1):
      S=0
      for j in np.arange(i+1,m):
        S+=A[i,j]*X[j,l]
      X[i,l]=(b[i,l]-S)/A[i,i]
  return X

def linsolve_triu_v00(A,b):
  assert is3Darray(A) 
  (N,m,n)=A.shape
  (getb,(Nb,mb,nb))=getScaComp(b,N)
  assert n==mb
  X=np.zeros((N,m,nb))
  for k in np.arange(N): 
    X[k]=linsolve_triu_sca(A[k],getb(b,k))
  return X

def linsolve_triu_v01(A,b):
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert n==m
  (getb,(Nb,mb,nb))=getComp(b,N)
  assert n==mb
  X=np.zeros((N,m,nb))
  S=np.zeros(N)
  for l in np.arange(nb): # Solve A[ X[:,l]=b[:,l]
    for i in np.arange(m-1,-1,-1):
      S*=0
      for j in np.arange(i+1,m):
        S+=A[:,i,j]*X[:,j,l]
      X[:,i,l]=(getb(b,i,l)-S)/A[:,i,i]
  return X

def linsolve_triu(A,b):
  return linsolve_triu_v01(A,b)
    
def linsolve_tril_sca(A,b):
  (m,n)=A.shape
  (mb,nb)=b.shape
  assert m==n and mb==m
  X=np.zeros((m,nb))
  for l in np.arange(nb): # Solve A X[:,l]=b[:,l]
    for i in np.arange(m):
      S=0
      for j in np.arange(i):
        S+=A[i,j]*X[j,l]
      X[i,l]=(b[i,l]-S)/A[i,i]
  return X

def linsolve_tril_v00(A,b):
  assert is3Darray(A) 
  (N,m,n)=A.shape
  (getb,(Nb,mb,nb))=getScaComp(b,N)
  assert n==mb
  X=np.zeros((N,m,nb))
  for k in np.arange(N): 
    X[k]=linsolve_tril_sca(A[k],getb(b,k))
  return X

def linsolve_tril_v01(A,b):
  assert is3Darray(A) 
  (N,m,n)=A.shape
  assert n==m
  (getb,(Nb,mb,nb))=getComp(b,N)
  assert n==mb
  X=np.zeros((N,m,nb))
  S=np.zeros(N)
  for l in np.arange(nb): # Solve A[ X[:,l]=b[:,l]
    for i in np.arange(m):
      S*=0
      for j in np.arange(i):
        S+=A[:,i,j]*X[:,j,l]
      X[:,i,l]=(getb(b,i,l)-S)/A[:,i,i]
  return X

def linsolve_tril(A,b):
  return linsolve_tril_v01(A,b)



def linsolve_cholesky_v00(A,b):
  L=fc_linalg3D.fact.cholesky_v00a(A)
  Y=linsolve_tril_v00(L,b)
  return linsolve_triu_v00(ctranspose(L),Y)

def linsolve_cholesky_v01(A,b):
  L=fc_linalg3D.fact.cholesky_v01(A)
  Y=linsolve_tril_v01(L,b)
  return linsolve_triu_v01(ctranspose(L),Y)

def linsolve_cholesky_old(A,b):
  return linsolve_cholesky_v01(A,b)

